
//==============================================================================
// L      CCCC M   M EEEEE TTTTT EEEEE RRRR         III  N   N OOOOO 
// L     C     MM MM E       T   E     R   R         I   NN  N O   O 
// L     C     M M M EEEE    T   EEEE  RRRR          I   N N N O   O 
// L     C     M M M E       T   E     R  R   ..     I   N  NN O   O 
// LLLLL  CCCC M   M EEEEE   T   EEEEE R   R  ..    III  N   N OOOOO 
//==============================================================================

//==============================================================================
// Program:      lcmeter.ino
// Author:       Pete Willard
// Version:      1.0d
// Target:       UNO R3, IDE 1.03 
// Date:         2013/04/24
// Time:         07:06:01
// Notes:        Latest Update : 10/11/2013  8:47AM

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

//  Uses Serial I/O
//  Uses LiquidCrystal Library
//  Uses Frequency Counter Library http://www.pjrc.com/teensy/td_libs.html
//  
// Reference:
// This design is heavily derivitive of Kerry Wong's LCF meter.
//  http://www.kerrywong.com/2010/10/16/avr-lc-meter-with-frequency-measurement/
// Some changes were made, but overall... it's mostly Kerry's design
//==============================================================================

//=====[ INCLUDE ]==============================================================

#include <LiquidCrystal.h>
#include <FreqCount.h>


//=====[ CONSTANTS ]============================================================
#define   DEBUG  0   // 0 = debugging disabled, 1 = enabled
enum STATE{
L = 1,C = 2};

// misc Settings
const int splashDelay      =  5 * 1000;     // Splash screen timer on boot
const int debounceDelay    = 10;      // milliseconds to wait until stable
const long int deviceTimeOut    = 300000;  // 5 minutes

// Frequency Settings, based on supplied parts
// Resonant Frequency: Mathematical Ideal--> 414473 HZ
// Actual value is based on some tweakage since OSC runs where it wants to.
const float Freq = 420000;          // Actual Frequency of Osc as measured
//                                  // With frequency meter  
const float C_ideal = 1800 * 1e-12; // ideal 1800 pF but we have a 2% part
const float L_ideal = 82  * 1e-6;   // ideal 82 uH but we have a 5% part 


//=====[ PINS ]=================================================================
//        Freq Pin  = 5
const int     mode_Pin  = 2;   // Input from relay mode pin
const int     set_C_Pin = 3;   // Relay Coil 1 - SET C mode
const int     set_L_Pin = 4;   // relay Coil 2 - SET L mode
// Switches
const int     cal_Switch  = 15; // Cabibrate SPST switch
const int     mode_Switch = 14; // Mode Select switch

const int Vbatt_Pin  = A3;  // Analog 3 is used to monitor LiPo battery
// LCD pins
const int pinRS = 12;      
const int pinEN = 11;
const int pinD4 = 10;
const int pinD5 = 9;
const int pinD6 = 8;
const int pinD7 = 7;
const int BACK_PIN = 6;   // Backlight Control - Note: just turn it on
const int PWR_PIN = 16;  // power off the device using Polulu switch
int LedPin = 13;
// LCD WIRING:
// LCD RS pin To digital pin 12
// LCD Enable pin To digital pin 11
// LCD D4 pin To digital pin 10
// LCD D5 pin To digital pin 9
// LCD D6 pin To digital pin 8
// LCD D7 pin To digital pin 7
// LCD R/W pin To ground
// 10K trimmer resistor:
// ends To +5V And ground
// wiper To LCD VO pin (pin 3)

// initialize the library With the numbers of the Interface pins
//                 RS  EN  D4 D5 D6 D7
LiquidCrystal lcd(pinRS, pinEN, pinD4, pinD5, pinD6, pinD7);


//=====[ VARIABLES ]============================================================

boolean backLightMode = false;
boolean ModeState = false;
boolean CalState = false;
boolean currentMode = false;
boolean turnOff = false;
boolean calFlag = false;


//  LC Timing
float L_value = L_ideal;    // Assign the ideal constants to variables
float C_value = C_ideal;    //
float L_offset = 0;         // For stray inducatance cancel in CAL routine
float C_offset = 0;         // For stray capaciatance cancel in CAL routine

unsigned long frq;          // Actual OSC frequency returned by library

long int previousMillis = 0; // For Auto-Off Timeout
long int elapsedMillis;
long int currentMillis;
//  Make sure we have a good battery, so know the voltage
float volts = 0.0; 
int rawdata = 0;
float vin = 0.0;
float ratio;    // Voltage Divider ratio
//=====[ SETUP ]================================================================
void setup() {
    Serial.begin(9600);
    pinMode(LedPin,OUTPUT);
    
    
    // Calibrate/Zero Switch
    pinMode(cal_Switch, INPUT);
    digitalWrite(cal_Switch, HIGH); // turn on pull-up resistors
    byte CalPreviousState = HIGH;
    
    // Mode Switch  - Used to change mode
    pinMode(mode_Switch, INPUT);
    digitalWrite(mode_Switch, HIGH);// turn on pull-up resistors
    byte ModePreviousState = HIGH;
    
    pinMode(PWR_PIN, OUTPUT);  // Automatic Power Down of Polulu Switch  
    digitalWrite(PWR_PIN, LOW);
    
    // Relay Control pins
    pinMode(set_L_Pin,OUTPUT);
    pinMode(set_C_Pin,OUTPUT);
    digitalWrite(set_L_Pin,LOW);
    digitalWrite(set_C_Pin,LOW);
    
    pinMode(mode_Pin,INPUT);  // to read the position of the latching relay
    digitalWrite(mode_Pin, HIGH);// turn on pull-up resistors
    
    // Backlight Control 
    pinMode(BACK_PIN, OUTPUT);  
    digitalWrite(BACK_PIN,LOW );
   
   // Initializers
    lcd_init(); 
    FreqCount.begin(1000);
}
//==============================================================================

//=====[ LOOP ]=================================================================
void loop() { 
    currentMillis = millis();
    
    if (FreqCount.available()) {  
        //   frequency available
        frq = FreqCount.read(); 
        checkLCMode();
        printLCMode();
        displayValue(frq); 
    }
    readSwitches();  
    checkBattery();
    
    // Autoshutoff
    if(currentMillis - previousMillis > deviceTimeOut) {
        digitalWrite(PWR_PIN, HIGH);
    }
    // If  readSwitches() noticed we wanted to calibrate, do it now.
    if (calFlag) {
        calibrate();
        calFlag = false;
    }
}


//==============================================================================
//=====[ SUBROUTINES ]==========================================================
void checkBattery() {
    // Battery Voltage check 
    rawdata = analogRead(A3);
    ratio = 2.08;  // relationship ratio between R1 ad R2 in voltage divider
    vin = rawdata * 0.00488; // Mv Steps per unit in ADC at 5V
    volts = vin * ratio; // ADC to  conversion
    if ((volts < 6.90) && (DEBUG == 0)) {
        lcd.clear();
        lcd.setCursor(0,0);
        
        lcd.print("Low Volt");
        lcd.setCursor(0,1);
        lcd.print("Pwr Down");
        delay(4000);
        digitalWrite(PWR_PIN, HIGH);
    }
}

//==============================================================================
void readSwitches() {
    if (debounce(cal_Switch)) {
        calFlag = true;
    }
    
    if (debounce(mode_Switch)) {
        if (currentMode == L) {
            PulsePin(set_C_Pin);
        }
        if (currentMode == C) {
            PulsePin(set_L_Pin);
        } 
    } 
    
}
//==============================================================================
void lcd_init() {
    // LCD initialization  
    lcd.begin(8, 2);  
    delay(1000);
    lcd.clear();
    // Print a splash message To the LCD.
    lcd.setCursor(0,0);
    lcd.print("LCMETER");
    lcd.setCursor(0,1);
    lcd.print("V 0.1d");
    delay(splashDelay);
    lcd.clear(); 
    lcd.setCursor(0,0); 
}

//==============================================================================
void printLCMode() {
    lcd.clear();
    lcd.setCursor(0,0);
    if (currentMode == L) {
        lcd.print("Coil ");
    } 
    if (currentMode == C) {
        lcd.print("Cap ");
        
    }
}

//==============================================================================
void checkLCMode() {
    if (digitalRead(mode_Pin) == LOW) {
        currentMode = L;
    } 
    else {
        currentMode = C;
    }  
}


//==============================================================================
// debounce returns true if the switch in the given pin is closed and stable
// From "Arduino Cookbook"
boolean debounce(int pin)
{
    boolean state;
    boolean previousState;
    
    previousState = digitalRead(pin); // store switch state
    for(int counter=0; counter < debounceDelay; counter++)
    {
        delay(1); // wait for 1 millisecond
        state = digitalRead(pin); // read the pin
        if( state != previousState)
        {
            counter = 0; // reset the counter if the state changes
            previousState = state; // and save the current state
        }   
    }
    
    // here when the switch state has been stable longer than the debounce period
    if(state == LOW) // LOW means pressed (because pull-ups are used)
    return true;
    else
    return false;
}

//==============================================================================
void TogglePin(char mypin) {
    digitalWrite(mypin,!digitalRead(mypin));
}
//==============================================================================
void PulsePin(char mypin) {
    TogglePin(mypin);
    delay(100);
    TogglePin(mypin);
} 

//==============================================================================
void calibrate() {
    // Generate an offset value for C  or L
    // Display running frequency
    // Show battery voltage
    
    lcd.clear();
    delay (20);
    lcd.setCursor(0,0);
    lcd.print("Cal/Zero");
    delay(1000);
    
    if (currentMode == L) {
        L_offset = Freq - frq;
        lcd.setCursor(0,0);
        lcd.print("Short   ");
        lcd.setCursor(0,1);
        lcd.print("Leads   ");
        delay(2000);
    } 
    if (currentMode == C) {
        C_offset = Freq - frq;
        lcd.setCursor(0,0);
        lcd.print("Open    ");
        lcd.setCursor(0,1);
        lcd.print("Leads   ");
        delay(2000);
    }
    
    lcd.clear();
    delay (20);
    lcd.setCursor(0,0);
    lcd.print("Cal/Zero");
    lcd.setCursor(0,1);
    lcd.print(frq);
    
    if (DEBUG) {
        Serial.println("Calibrate - Zero");
        Serial.print("Frequency ");
        Serial.println(frq);
        Serial.print(" LOFFSET ");
        Serial.println(L_offset);
        Serial.print(" COFFSET ");
        Serial.println(C_offset);
    }
    
    delay(2000);
    lcd.clear();
    delay (20); 
    lcd.setCursor(0,0);
    lcd.print("VOLTAGE");
    lcd.setCursor(0,1);
    lcd.print(volts);
    delay(splashDelay);
    lcd.clear(); 
    lcd.setCursor(0,0); 
    
}

//==============================================================================
float calcValue(float f, float VRef) {
    float v = 0;
    v = abs(((Freq * Freq) / (f * f) - 1.0) * VRef);
    return v;
}
//==============================================================================
//*******************************************************************************
void displayValue(long fin) {
    float f = (float) fin;
    float v = 0;
    float C_measure;
    float L_measure;
    C_measure = f + C_offset; 
    L_measure = f + L_offset;
    
    lcd.setCursor(0, 1);
    lcd.print("        ");
    lcd.setCursor(0, 1);
    
    switch (currentMode) {
        case C:
        
        v = calcValue(C_measure,C_value);
        
        if (v < 1e-9) {
            v = v * 1e12; // pico
            lcd.print(v);
            lcd.print(" ");
            lcd.setCursor(6,1);
            lcd.print("pF");
            
        } 
        else if (v >= 1e-9 && v < 1e-8) {
            v = v * 1e9; // n
            lcd.print(v);
            lcd.print(" ");
            lcd.setCursor(6,1);
            lcd.print("nF");
        }
        else if (v >= 1e-8 && v < 1e-4) {
            v = v * 1e6; // n
            lcd.print(v);
            lcd.print(" ");
            lcd.setCursor(6,1);
            lcd.print("uF");
        }
        else {
            lcd.print("---");
            if (DEBUG) {
                Serial.print("---");
            }
        }
        break;
        case L:
        v = calcValue(L_measure,L_value);
        
        if (v < 1e-6) {
            v = v * 1e9; //nH
            lcd.print(v);
            lcd.print(" ");
            lcd.setCursor(6,1);
            lcd.print("nH");
            if (DEBUG) {
                Serial.print(v);
                Serial.print(" ");
                Serial.print("nH");
            }
        } 
        else if (v >= 1e-6 && v < 1e-3) {
            v = v * 1e6; //uH
            lcd.print(v);
            lcd.print(" ");
            lcd.setCursor(6,1);
            lcd.print("uH");
            if (DEBUG) {
                Serial.print(v);
                Serial.print(" ");
                Serial.print("uH");
            }
        } 
        else if (v >= 1e-3 && v < 1) {
            v = v * 1e3;
            lcd.print(v);
            lcd.print(" ");
            lcd.setCursor(6,1);
            lcd.print("mH");
            if (DEBUG) {
                Serial.print(v);
                Serial.print(" ");
                Serial.print("mH");
            }
        } 
        else if (v >= 1 && v < 100) {
            lcd.print(v);
            lcd.print(" ");
            lcd.setCursor(6,1);
            lcd.print("H");
            if (DEBUG) {
                Serial.print(v);
                Serial.print(" ");
                Serial.print("H");
            }
        } 
        else {
            lcd.print("---");
            if (DEBUG) {
                Serial.print("---"); 
            }
            
        }
        break;
    }
    
}







